/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2012                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/
/*! \file trimesh_base.cpp
\ingroup code_sample

\brief the minimal example of using the lib

This file contain a minimal example of the library, showing how to load a mesh and how to compute per vertex normals on it.

*/

#include<vcg/complex/complex.h>
#include<wrap/io_trimesh/import_off.h>
#include<wrap/io_trimesh/export_off.h>
#include<wrap/io_trimesh/export_obj.h>
#include<wrap/io_trimesh/export_ply.h>
#include<vcg/complex/algorithms/update/topology.h>
#include<vcg/complex/algorithms/update/normal.h>
#include<vcg/complex/algorithms/update/color.h>
#include<vcg/complex/algorithms/clean.h>
#include <vcg/space/index/grid_static_ptr.h>
#include <vcg/space/index/kdtree/kdtree.h>
#include<vcg/space/distance3.h>

class MyVertex; class MyEdge; class MyFace;
struct MyUsedTypes : public vcg::UsedTypes<vcg::Use<MyVertex>   ::AsVertexType,
                                           vcg::Use<MyEdge>     ::AsEdgeType,
                                           vcg::Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public vcg::Vertex< MyUsedTypes, vcg::vertex::Coord3f, vcg::vertex::Normal3f, vcg::vertex::BitFlags, vcg::vertex::Color4b, vcg::vertex::VEAdj, vcg::vertex::Qualityf  >{};
class MyFace    : public vcg::Face<   MyUsedTypes, vcg::face::FFAdj,  vcg::face::VertexRef, vcg::face::Color4b, vcg::face::Normal3f, vcg::face::Mark, vcg::face::BitFlags > {};
class MyEdge    : public vcg::Edge<   MyUsedTypes, vcg::edge::VertexRef, vcg::edge::VEAdj, vcg::edge::EEAdj, vcg::edge::BitFlags> {};

class MyMesh    : public vcg::tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace> , std::vector<MyEdge>  > {};
using namespace vcg;



int countNonVisitedEdges(MyVertex * vp, MyEdge * &ep)
{
  std::vector<MyEdge *> starVec;
  edge::VEStarVE(&*vp,starVec);

  int cnt =0;
  for(size_t i=0;i<starVec.size();++i)
  {
    if(!starVec[i]->IsV())
    {
      cnt++;
      ep = starVec[i];
    }
  }

  return cnt;
}

void Retract(MyMesh &t)
{
  tri::Clean<MyMesh>::RemoveDuplicateVertex(t);
  printf("Retracting a mesh of %i edges and %i vertices\n",t.en,t.vn);
  tri::UpdateTopology<MyMesh>::VertexEdge(t);

  std::stack<MyVertex *> vertStack;

  for(MyMesh::VertexIterator vi=t.vert.begin();vi!=t.vert.end();++vi)
  {
    std::vector<MyEdge *> starVec;
    edge::VEStarVE(&*vi,starVec);
    if(starVec.size()==1)
      vertStack.push(&*vi);
  }

  tri::UpdateFlags<MyMesh>::EdgeClearV(t);
  tri::UpdateFlags<MyMesh>::VertexClearV(t);

  while(!vertStack.empty())
  {
    MyVertex *vp = vertStack.top();
    vertStack.pop();
    vp->C()=Color4b::Blue;
    MyEdge *ep=0;

    int eCnt =  countNonVisitedEdges(vp,ep);
    if(eCnt==1) // We have only one non visited edge over vp
    {
      assert(!ep->IsV());
      ep->SetV();
      MyVertex *otherVertP;
      if(ep->V(0)==vp) otherVertP = ep->V(1);
      else otherVertP = ep->V(0);
      vertStack.push(otherVertP);
    }
  }

  tri::io::ExporterPLY<MyMesh>::Save(t,"retract0.ply",tri::io::Mask::IOM_EDGEINDEX | tri::io::Mask::IOM_VERTCOLOR);

  for(size_t i =0; i<t.edge.size();++i)
    if(t.edge[i].IsV()) tri::Allocator<MyMesh>::DeleteEdge(t,t.edge[i]);

  tri::Clean<MyMesh>::RemoveUnreferencedVertex(t);

  tri::Allocator<MyMesh>::CompactEveryVector(t);
  tri::io::ExporterPLY<MyMesh>::Save(t,"retract1.ply",tri::io::Mask::IOM_EDGEINDEX | tri::io::Mask::IOM_VERTCOLOR);

  for(size_t i =0; i<t.edge.size();++i)
    if(t.edge[i].IsV()) tri::Allocator<MyMesh>::DeleteEdge(t,t.edge[i]);

  tri::Allocator<MyMesh>::CompactEveryVector(t);
}
typedef vcg::GridStaticPtr<MyMesh::FaceType, MyMesh::VertexType::ScalarType> MeshGrid;

void Smooth(MyMesh &t, MyMesh &base)
{
  // Construction of the uniform grid
  MeshGrid uniformGrid;
  uniformGrid.Set(base.face.begin(), base.face.end());
  tri::UpdateColor<MyMesh>::PerVertexConstant(t,Color4b::White);
  tri::UpdateNormal<MyMesh>::PerFaceNormalized(base);
  for(int k=0;k<200;++k)
  {
    std::vector<Point3f> posVec(t.vn,Point3f(0,0,0));
    std::vector<int> cntVec(t.vn,0);

    for(size_t i =0; i<t.en;++i)
      for(int j=0;j<2;++j)
      {
        int vertInd = tri::Index(t,t.edge[i].V(j));
        posVec[vertInd] += t.edge[i].V1(j)->P();
        cntVec[vertInd] ++;
      }

    const float maxDist = base.bbox.Diag()/10.0;
    for(size_t i =0; i<t.vn;++i)
    {
      Point3f newP = (t.vert[i].P() + posVec[i])/float(cntVec[i]+1);
      float minDist;
      Point3f closestP,ip;
      MyFace *f = vcg::tri::GetClosestFaceBase(base,uniformGrid,newP,maxDist, minDist, closestP);
      assert(f);
      t.vert[i].P() = closestP;
    }
  }
  tri::io::ExporterPLY<MyMesh>::Save(t,"retract2.ply",tri::io::Mask::IOM_EDGEINDEX | tri::io::Mask::IOM_VERTCOLOR);
}

void ComputeField(MyMesh &t, MyMesh &base)
{
  tri::UpdateNormal<MyMesh>::PerVertexNormalized(base);
  vcg::ConstDataWrapper<MyMesh::VertexType::CoordType> wrapperVcg(&t.vert[0].P(), t.vert.size(), size_t(t.vert[1].P().V()) - size_t(t.vert[0].P().V()));
  vcg::KdTree<MyMesh::ScalarType> kdTree(wrapperVcg);
  tri::UpdateTopology<MyMesh>::VertexEdge(t);
  
  std::vector<MyEdge *> closestEdgeToVertVec(base.vn);

  for(MyMesh::VertexIterator vi=base.vert.begin();vi!=base.vert.end();++vi)
  {
    const Point3f p = vi->P();
    vcg::KdTree<MyMesh::ScalarType>::PriorityQueue queue;
    kdTree.doQueryK(p, 2, queue);
    int nearestInd = queue.getIndex(0);
    MyVertex * vp0 =  &(t.vert[nearestInd]);
    MyEdge * ep = vp0->VEp();
    int      ei = vp0->VEi();
    vi->Q() = std::numeric_limits<float>::max();
    MyEdge *cep=0;
    while(ep)
    {
      assert(ep->V(ei) == vp0 );
      Segment3f seg(ep->V0(ei)->P(),ep->V1(ei)->P());
      Point3f close;
      float dist;
      vcg::SegmentPointDistance(seg,p,close,dist);

      if(vi->Q() > dist)
      {
        vi->Q() = dist;
        cep = ep;
      }
      MyEdge *newEp= ep->VEp(ei);
      int newEi= ep->VEi(ei);
      ep=newEp;
      ei=newEi;
    }
    closestEdgeToVertVec[tri::Index(base,*vi)] = cep;

    Point3f nn = vcg::Normal(p,cep->V(0)->P(),cep->V(1)->P());
    if(nn * vi->N() < 0 ) vi->Q() = -vi->Q();
  }

  float eps = base.bbox.Diag()/100.0;
  tri::UpdateColor<MyMesh>::PerVertexQualityRamp(base);
  tri::io::ExporterPLY<MyMesh>::Save(base,"distField.ply",tri::io::Mask::IOM_VERTQUALITY | tri::io::Mask::IOM_VERTCOLOR);
  for(int i =0 ;i<base.fn;++i)
  {
    Segment3f segV[3];
    for(int j=0;j<3;++j)
    { 
      MyEdge *ep = closestEdgeToVertVec[tri::Index(base,base.face[i].V(j))];
      segV[j] = Segment3f(ep->V(0)->P(),ep->V(1)->P());
    }   
    
    for(int j=0;j<3;++j)      
    {
      bool parallelFlag; Point3f clos0,clos1;
      float dist;
      MyVertex *v0 = base.face[i].V0(j);
      MyVertex *v1 = base.face[i].V1(j);
      if(v0->Q() * v1->Q() < 0)
      {
        Segment3f segTri(v0->P(),v1->P());
        for(int k=0;k<3;++k)
        {
        SegmentSegmentDistance(segTri,segV[k],dist,parallelFlag,clos0,clos1);
        if(dist< eps)        
          base.face[i].C() = Color4b::LightBlue;                  
        }
      }
      
    }

  }
  tri::io::ExporterPLY<MyMesh>::Save(base,"faceCrossing.ply",tri::io::Mask::IOM_VERTQUALITY | tri::io::Mask::IOM_FACECOLOR);
  
}


/*
 * Make an edge mesh 1-manifold by splitting all the
 * vertexes that have more than two incident edges
 * 
 * It performs the split in three steps. 
 * First it collect and count the vertices to be splitten. 
 * Then add the vertices to the mesh and lastly update the newly added vertices. 
 * 
 */

void DecomposeNonManifoldTree(MyMesh &m)
{
  std::vector<int> degreeVec(m.vn);
  tri::UpdateTopology<MyMesh>::VertexEdge(m);
  int neededVert=0;
  for(MyMesh::VertexIterator vi=m.vert.begin(); vi!=m.vert.end();++vi)
  {
    std::vector<MyEdge *> starVec;
    edge::VEStarVE(&*vi,starVec);
    degreeVec[tri::Index(m, *vi)] = starVec.size();
    if(starVec.size()>2)
      neededVert += starVec.size()-2;
  }

  MyMesh::VertexIterator firstVi = tri::Allocator<MyMesh>::AddVertices(m,neededVert);

  for(int i=0;i<degreeVec.size();++i)
  {
    if(degreeVec[i]>2)
    {
      std::vector<MyEdge *> starVec;
      edge::VEStarVE(&(m.vert[i]),starVec);
      assert(starVec.size() == degreeVec[i]);
      for(int j=2;j<starVec.size();++j)
      {
        MyEdge *ep = starVec[j];
        int ind; // index of the vertex to be changed
        if(tri::Index(m,ep->V(0)) == i) ind = 0;
            else ind = 1;

        ep->V(ind) = &*firstVi;
        ep->V(ind)->P() = m.vert[i].P();
        ++firstVi;
      }
    }
  }
  tri::io::ExporterPLY<MyMesh>::Save(m,"retract3.ply",tri::io::Mask::IOM_EDGEINDEX);
}

/*
 * Given a manifold edgemesh it returns the boundary/terminal endpoints.
 */
void FindTerminalPoints(MyMesh &m,std::vector<MyVertex *> &vec)
{
  tri::UpdateTopology<MyMesh>::VertexEdge(m);
  for(MyMesh::VertexIterator vi=m.vert.begin(); vi!=m.vert.end();++vi)
  {
    std::vector<MyEdge *> starVec;
    edge::VEStarVE(&*vi,starVec);
    if(starVec.size()==1)
     vec.push_back(&*vi);
  }
}


/*
 * Given a mesh and vector of vertex pointers it splits the mesh with the given points.
 * To avoid degeneracies it snaps the splitting points that came nearest than a given
 * threshold to the edge or the vertex of the closest triangle. When an input vertex
 * is snapped the passed vertex position is modiried accordingly
 * (this is the reason for having a vector of vertex pointers instead just a vector of points)
 *
 */
void SplitMeshWithPoints(MyMesh &m, std::vector<MyVertex *> &vec)
{
  printf("Splitting with %i vertices\n",vec.size());
  int faceToAdd=0;
  int vertToAdd=0;

  // For each splitting point we save the index of the face to be splitten and the "kind" of split to do:
  // 3 -> means classical 1 to 3 face split
  // 2 -> means edge split.
  // 0 -> means no need of a split (e.g. the point is coincident with a vertex)

  std::vector< std::pair<int,int> > toSplitVec(vec.size(), std::make_pair(0,0));
  MeshGrid uniformGrid;
  uniformGrid.Set(m.face.begin(), m.face.end());
  const float eps=0.01f;
  const float maxDist = m.bbox.Diag()/10.0;

  for(size_t i =0; i<vec.size();++i)
  {
    Point3f newP = vec[i]->P();
    float minDist;
    Point3f closestP,ip;
    MyFace *f = vcg::tri::GetClosestFaceBase(m,uniformGrid,newP,maxDist, minDist, closestP);
    assert(f);

//    if(ip[0]<eps || ip[1]<eps || ip[2]<eps)
//    {
//      // TODO
//    }
//    else
    {
      toSplitVec[i].first = tri::Index(m,f);
      toSplitVec[i].second = 3;
      faceToAdd += 2;
      vertToAdd += 1;
    }
  }
  printf("adding %i faces and %i vertices\n",faceToAdd,vertToAdd);
  MyMesh::FaceIterator newFi = tri::Allocator<MyMesh>::AddFaces(m,faceToAdd);
  MyMesh::VertexIterator newVi = tri::Allocator<MyMesh>::AddVertices(m,vertToAdd);

  tri::UpdateColor<MyMesh>::PerFaceConstant(m,Color4b::White);

  for(size_t i =0; i<vec.size();++i)
  {
    if(toSplitVec[i].second==3)
    {
     MyFace *fp0,*fp1,*fp2;
     fp0 = &m.face[toSplitVec[i].first];
     fp1 = &*newFi; newFi++;
     fp2 = &*newFi; newFi++;
     MyVertex *vp = &*(newVi++);
     vp->P() = vec[i]->P();
     MyVertex *vp0 = fp0->V(0);
     MyVertex *vp1 = fp0->V(1);
     MyVertex *vp2 = fp0->V(2);

     fp0->V(0) = vp0; fp0->V(1) = vp1; fp0->V(2) = vp;
     fp1->V(0) = vp1; fp1->V(1) = vp2; fp1->V(2) = vp;
     fp2->V(0) = vp2; fp2->V(1) = vp0; fp2->V(2) = vp;

     fp0->C() = Color4b::Green;
     fp1->C() = Color4b::Green;
     fp2->C() = Color4b::Green;
    }
  }

  tri::io::ExporterOFF<MyMesh>::Save(m,"split.off",tri::io::Mask::IOM_FACECOLOR);
}

// This function will decompose the input edge mesh into a set of 
// connected components.
// the vector will contain the
void BuildConnectedComponentVectors(MyMesh &m, std::vector< std::vector< int> > &ccVec)
{
  tri::UpdateFlags<MyMesh>::EdgeClearV(m);
  tri::UpdateTopology<MyMesh>::EdgeEdge(m);

  int visitedEdgeNum=0 ;
  int ccCnt=0;
  
  MyMesh::EdgeIterator eIt = m.edge.begin();  
  while(visitedEdgeNum < m.en)
  {
    ccVec.resize(ccVec.size()+1);
    while(eIt->IsV()) ++eIt;
    MyEdge *startE = &*eIt;
    
    MyEdge *curEp = &*eIt;
    int     curEi = 0;
    printf("Starting Visit of connected Component %i from edge %i\n",ccCnt,tri::Index(m,*eIt));
    while( (curEp->EEp(curEi) != startE) &&
           (curEp->EEp(curEi) != curEp) )
    {
      MyEdge *nextEp = curEp->EEp(curEi);
      int nextEi =  (curEp->EEi(curEi)+1)%2;
      curEp = nextEp;
      curEi = nextEi;
    }   

    curEp->SetV();
    curEi = (curEi +1)%2; // Flip the visit direction!
    visitedEdgeNum++;
    ccVec[ccCnt].push_back(tri::Index(m,curEp));        
    while(! curEp->EEp(curEi)->IsV())
    {
      visitedEdgeNum++;      
      MyEdge *nextEp = curEp->EEp(curEi);
      int nextEi =  (curEp->EEi(curEi)+1)%2;
      curEp->SetV();
      curEp = nextEp;
      curEi = nextEi;
      curEp->V(curEi)->C() = Color4b::Scatter(10,ccCnt);
      ccVec[ccCnt].push_back(tri::Index(m,curEp));      
    }
    ccCnt++;    
  }
}

void ExtractSubMesh(MyMesh &T, std::vector<int> ind, MyMesh &sub)
{
  sub.Clear();
  std::vector<int> remap(T.vert.size(),-1);
  for(int i=0;i<ind.size();++i)
  {
    int v0 = tri::Index(T,T.edge[ind[i]].V(0));
    int v1 = tri::Index(T,T.edge[ind[i]].V(1));
    if(remap[v0]=-1) 
    {
      remap[v0]=sub.vn;
      tri::Allocator<MyMesh>::AddVertex(sub,T.vert[v0].P());      
    }
    if(remap[v1]=-1) 
    {
      remap[v1]=sub.vn;
      tri::Allocator<MyMesh>::AddVertex(sub,T.vert[v1].P());      
    }
    tri::Allocator<MyMesh>::AddEdge(sub, &sub.vert[remap[v0]], &sub.vert[remap[v1]]);   
  }  
}

void Reorient(MyMesh &m, std::vector< std::vector< int> > &ccVec)
{
  for(int i=0;i<ccVec.size();++i)
  {
    std::vector<bool> toFlipVec(ccVec[i].size(),false);
    
    for(int j=0;j<ccVec[i].size();++j)
    {
      MyEdge *cur  = & m.edge[ccVec[i][j]];
      MyEdge *prev;
      if(j==0) prev = cur;
          else prev = & m.edge[ccVec[i][j-1]];
      
      if(cur->EEp(0) != prev)
      {
        toFlipVec[j] = true;
        assert(cur->EEp(1) == prev);
      }      
    }
    for(int j=1;j<ccVec[i].size();++j)
      if(toFlipVec[j]) 
        std::swap(m.edge[ccVec[i][j]].V(0), m.edge[ccVec[i][j]].V(1));    
  }
}

int main( int argc, char **argv )
{
  if(argc<2)
  {
    printf("Usage trimesh_base <meshfilename.off>\n");
    return -1;
  }
  MyMesh m;
  tri::io::ImporterOFF<MyMesh>::Open(m,argv[1]);
  printf( "Mesh has %i vert and %i faces\n", m.VN(), m.FN() );
  tri::UpdateTopology<MyMesh>::FaceFace(m);

  tri::UpdateFlags<MyMesh>::FaceClearV(m);

  std::vector<face::Pos<MyFace> > visitStack;

  typedef face::Pos<MyFace> MyPos;
  MyMesh treeMesh, dualMesh;
  m.face[0].SetV();
  for(int i=0;i<3;++i)
    visitStack.push_back(MyPos(&(m.face[0]),i,m.face[0].V(i)));

  int cnt=1;

  while(!visitStack.empty())
  {
    std::swap(visitStack.back(),visitStack[rand()%visitStack.size()]);
    MyPos c = visitStack.back();
    visitStack.pop_back();
    assert(c.F()->IsV());
    c.F()->C() = Color4b::ColorRamp(0,m.fn,cnt);
    c.FlipF();
    if(!c.F()->IsV())
    {
      tri::Allocator<MyMesh>::AddEdge(treeMesh,Barycenter(*(c.FFlip())),Barycenter(*(c.F())));
      ++cnt;
      c.F()->SetV();
      c.FlipE();c.FlipV();
      visitStack.push_back(c);
      c.FlipE();c.FlipV();
      visitStack.push_back(c);
    }
    else
    {
      tri::Allocator<MyMesh>::AddEdge(dualMesh,c.V()->P(),c.VFlip()->P());
    }
  }
  assert(cnt==m.fn);

  tri::io::ExporterOFF<MyMesh>::Save(m,"out.off",tri::io::Mask::IOM_FACECOLOR);
  tri::io::ExporterPLY<MyMesh>::Save(treeMesh,"tree.ply",tri::io::Mask::IOM_EDGEINDEX);
  tri::io::ExporterPLY<MyMesh>::Save(dualMesh,"dual.ply",tri::io::Mask::IOM_EDGEINDEX);

  Retract(dualMesh);
  Smooth(dualMesh,m);
  DecomposeNonManifoldTree(dualMesh);
  std::vector< std::vector<int> > ccVec;
  BuildConnectedComponentVectors(dualMesh, ccVec);
  tri::io::ExporterPLY<MyMesh>::Save(dualMesh,"coloredComponent.ply",tri::io::Mask::IOM_EDGEINDEX+tri::io::Mask::IOM_VERTCOLOR);
  Reorient(dualMesh,ccVec);
  std::vector<MyVertex *> terminalVec;
  FindTerminalPoints(dualMesh,terminalVec);
  SplitMeshWithPoints(m,terminalVec);
  MyMesh sub;
  ExtractSubMesh(dualMesh,ccVec[0],sub);
  ComputeField(sub,m);
  return 0;
}









